package main;

import java.util.*;

import java.awt.*;
import java.awt.image.*;

public class Dust {
	Vector2D pos;
	Vector2D vel;
	Vector2D intersect;
	
	Vector2D repulse;
	int hp;
	
	Color color;
	
	Dust(double posX, double posY, double velX, double velY) {
		pos = new Vector2D(posX, posY);
		vel = new Vector2D(velX, velY);
		hp = 255;
		color = Color.black;
	}
	
}

class DustCloud {
	ArrayList<Dust> flock;
	double redMod, blueMod, greenMod;
	
	DustCloud(int size, double centX, double centY, double radius, double velX, double velY, double redMod, double greenMod, double blueMod) {
		flock = new ArrayList<Dust>();
		for(int i = 0; i < size; i++) {
			Dust newDust = new Dust(centX + radius * (Math.random() * 2 - 1.0), centY + radius * (Math.random() * 2 - 1.0), velX, velY);
			flock.add(newDust);
		}
		
		this.redMod = redMod;
		this.greenMod = greenMod;
		this.blueMod = blueMod;
	}
	
	Vector2D flockCenter() {
		double x = 0.0;
		double y = 0.0;
		for(int i = 0; i < flock.size(); i++) {
			Dust curDust = (Dust) flock.get(i);
			x += curDust.pos.x;
			y += curDust.pos.y;
			 
		}
		x /= flock.size();
		y /= flock.size();
		
		return new Vector2D(x, y);
	}
	
	Vector2D flockVelocity() {
		double x = 0.0;
		double y = 0.0;
		for(int i = 0; i < flock.size(); i++) {
			Dust curDust = (Dust) flock.get(i);
			x += curDust.vel.x;
			y += curDust.vel.y;
		}
		x /= flock.size();
		y /= flock.size();
		
		return new Vector2D(x, y);
	}
	
	public void moveCloud(int targX, int targY) {
		//((Dust) flock.get(0)).pos = new Vector2D(targX, targY);
		
		Vector2D center = flockCenter();
		Vector2D avVel = flockVelocity();
		
		for(Dust curDust : flock) {
			//Dust curDust = (Dust) flock.get(i);
			
			// Steer towards center of flock
			Vector2D toCenter = (new Vector2D(center.x - curDust.pos.x, center.y - curDust.pos.y));
			//toCenter.add(new Vector2D(targX - curDust.pos.x, targY - curDust.pos.y));
			toCenter.scale(0.01);
			curDust.vel.add(toCenter);
			
			// Steer in direction of flock
			Vector2D toVel = (new Vector2D(curDust.vel.x - avVel.x, curDust.vel.y - avVel.y));
			toVel.scale(0.01);
			curDust.vel.add(toVel);
			
			// Steer to avoid neighbors
			Vector2D toNeighbors = new Vector2D(0.0, 0.0);
			for(Dust curNeighbor : flock) {
				if(curDust != curNeighbor) {
					Vector2D toNeighbor = (new Vector2D(curDust.pos.x - curNeighbor.pos.x, curDust.pos.y - curNeighbor.pos.y));
					
						
					if(toNeighbor.length() < 15) {
						toNeighbor.scale(3.0 / toNeighbor.length());
					} else if (toNeighbor.length() < 45) {
						toNeighbor.scale(1.5 / toNeighbor.length());
					} else {
						toNeighbor.scale(0.2 / toNeighbor.length());
					}
						
					toNeighbors.add(toNeighbor);
					
				}
			}
			
			
			toNeighbors.scale(0.03);
			
			// Steer to avoid obstacles
			Vector2D toObstacles = new Vector2D(0.0, 0.0);
			for(Wall curWall : Wall.allWall) {
				toObstacles.add(curWall.wallSeg.getRepulsion(curDust));
			}
			
			Vector2D downWall = new Vector2D(0.0,0.0);
			LineSeg dustToBase = new LineSeg(curDust.pos.x , curDust.pos.y, RenderFrame.pubFrame.mainBase.posX, RenderFrame.pubFrame.mainBase.posY);
			double maxDist = dustToBase.getLength();
			if(maxDist < 200){
				for(Wall curWall : Wall.allWall){
					Vector2D intersect = dustToBase.getIntersect(curWall.wallSeg);
					if(intersect != null) {
						Vector2D intClone = new Vector2D(intersect.x - dustToBase.end1X, intersect.y - dustToBase.end1Y);
						if(intClone.length() < maxDist) {
							downWall.x = curWall.wallSeg.end2X - curWall.wallSeg.end1X;
							downWall.y = curWall.wallSeg.end2Y - curWall.wallSeg.end1Y;
							maxDist = intClone.length();
						}
					}
				}
			}			
			
			downWall.scale(0.009);			
			
			if(curDust.hp < 255)
				curDust.hp++;
			
			for(int i = 0; i < Light.allLight.size(); i++) {
				if(i < Light.allLight.size()){
					Light curLight = Light.allLight.get(i);
					if(curLight != null && curLight.lightSeg != null) {
					LineSeg occludeLight = new LineSeg(curLight.lightSeg.end1X, curLight.lightSeg.end1Y,
							curLight.lightSeg.end1X + curLight.maxDist * Math.cos(curLight.lightSeg.getTheta()),
							curLight.lightSeg.end1Y + curLight.maxDist * Math.sin(curLight.lightSeg.getTheta()));
					
					Vector2D repulse = occludeLight.getRepulsion(curDust);
					if(curLight.leftLightSeg != null && curLight.rightLightSeg != null) {
						Vector2D repulseLeft = curLight.leftLightSeg.getRepulsion(curDust);
						Vector2D repulseRight = curLight.rightLightSeg.getRepulsion(curDust);
						repulse.add(repulseLeft);
						repulse.add(repulseRight);
						repulse.scale(1.0 / 3.0);
					} 
					
					if(repulse.length() > 0.6)
						curDust.hp-=5;
					repulse.scale(0.05);
					toObstacles.add(repulse);
					}
				}
			}
			toObstacles.scale(2.0);
			curDust.repulse = toObstacles;
			
			Vector2D toBase = new Vector2D((RenderFrame.pubFrame.mainBase.posX - curDust.pos.x), (RenderFrame.pubFrame.mainBase.posY - curDust.pos.y));
			
			
			
			double dist = toBase.length();
			double damage = 0.15 / dist;
			if(damage > 0.5)
				damage = 0.5;
			if(damage < 0.01)
				damage = 0.0;
			RenderFrame.pubFrame.mainBase.damagePoints+=damage;
			
			
			toBase.normalize();
			toBase.scale(0.01);
			
			curDust.vel.add(toBase);
			curDust.vel.add(toNeighbors);
			curDust.vel.add(new Vector2D (Math.random() * 0.01 - 0.005, Math.random() * 0.01 - 0.005));
			curDust.vel.add(toObstacles);
			curDust.vel.add(downWall);
			if(curDust.vel.length() > 4.0) {
				curDust.vel.scale(0.80);
				if(curDust.vel.length() > 10.0)
					curDust.vel.scale(0.10);			
			}			
			
			
		}
		
		// Update position separately, so it doesn't influence neighbor-avoidance
		for(Dust curDust : flock) {
			curDust.pos.add(curDust.vel);
		}
		
		// Remove any dead dust
		for(int i = 0; i < flock.size(); i++) {
			Dust curDust = flock.get(i);
			if(curDust.hp < 1) {
				flock.remove(curDust);
				i--;
			}
		}
		
		
	}
	
	public void paint(Graphics g) {
		if (RenderFrame.phase == 1) //only display the dust if it is night - not the best solution...
		{
			for(int i = 0; i < flock.size(); i++) {
				Dust curDust = (Dust) flock.get(i);
			
				BufferedImage dustImage = new BufferedImage(32, 32, BufferedImage.TYPE_INT_ARGB);
			
				Graphics2D dustG2D = (Graphics2D) dustImage.getGraphics();
				dustG2D.drawImage(RenderFrame.mote, 0, 0, RenderFrame.pubFrame);
				dustG2D.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_IN, 1.0f));
			
				int blueCol = (int) ((255 - curDust.hp) * blueMod);
				if(blueCol < 0)
					blueCol = 0;
				if(blueCol > 255)
					blueCol = 255;
			
				int redCol = (int) ((255 - curDust.hp) * redMod);
				if(redCol < 0)
					redCol = 0;
				if(redCol > 255)
					redCol = 255;
			
				int greenCol = (int) ((255 - curDust.hp) * greenMod);
				if(greenCol < 0)
					greenCol = 0;
				if(greenCol > 255)
					greenCol = 255;
			
				Color dustCol = new Color(redCol, greenCol, blueCol, 255);
			
				dustG2D.setColor(dustCol);
				for(int ii = 0; ii < 32; ii++) {
					dustG2D.drawLine(0, ii, 32, ii);
				}
			
				g.drawImage(dustImage, (int) curDust.pos.x, (int) curDust.pos.y, RenderFrame.pubFrame);
			
			
			}
		}
	}
	
}

class Vector2D {
	double x, y;
	Vector2D(double x, double y) {
		this.x = x;
		this.y = y;
	}
	
	void add(Vector2D toAdd) {
		this.x += toAdd.x;
		this.y += toAdd.y;
	}
	
	void scale(double scalar) {
		this.x *= scalar;
		this.y *= scalar;
	}
	
	void normalize() {
		double len = length();
		this.x *= (1 / len);
		this.y *= (1 / len);
	}
	
	double theta() {
		return Math.atan2(y, x);
	}
	
	void rotate(double t) {
		double newX = length() * Math.cos(theta() + t);
		double newY = length() * Math.sin(theta() + t);
		this.x = newX;
		this.y = newY;
	}
	
	double length() {
		return Math.sqrt(x * x + y * y);
	}
}