package main;

import java.util.*;
import java.awt.*;

public class Light {
	static ArrayList<Light> allLight = new ArrayList<Light>();
	static boolean lightTrace = true;
	boolean reflectChecked;
	double maxDist;
	
	int recurse;
	LineSeg lightSeg;
	LineSeg rightLightSeg; // Line segment that underlies the light beam - should be replaced w/left and right bound segments
	LineSeg leftLightSeg;
	LineSeg baseSeg;
	public Light (LineSeg lightSeg, int recurse) {
		this.lightSeg = lightSeg;				
		this.recurse = recurse;
		reflectChecked = false;
		this.baseSeg = new LineSeg(lightSeg.end1X, lightSeg.end1Y, lightSeg.end1X, lightSeg.end1Y);
		allLight.add(this); // Add this instance to the overall collection of lights on creation
	}
	
	public Light(LineSeg lightSeg, LineSeg baseSeg, int recurse) {
		this.lightSeg = lightSeg;			
		this.recurse = recurse;
		reflectChecked = false;
		this.baseSeg = baseSeg;
		maxDist = lightSeg.getLength();
		allLight.add(this); // Add this instance to the overall collection of lights on creation
	}
	
	public void rotate() {
		Vector2D toRot = new Vector2D(lightSeg.end2X - lightSeg.end1X, lightSeg.end2Y - lightSeg.end1Y);
		toRot.rotate(0.01);
		lightSeg.end2X = lightSeg.end1X + toRot.x;
		lightSeg.end2Y = lightSeg.end1Y + toRot.y;
		maxDist = 0.0;
		rightLightSeg = lightSeg;
		leftLightSeg = lightSeg;
	}
	
	public boolean reflect() {
		boolean reflected = false;
		if(this.recurse < 20) {
			reflectChecked = true;
			Vector2D reflect = new Vector2D(0,0);
			Vector2D toReflect = new Vector2D(0,0);
			LineSeg mirror = null;
			double maxReflect = maxDist + 2;//lightSeg.getLength();
			for(int i = 0; i < Mirror.allMirror.size(); i++) {
				Mirror curMirror = Mirror.allMirror.get(i);
				Vector2D temp = lightSeg.getIntersect(curMirror.mirrorSeg);			
				if(temp != null){
					Vector2D refClone = new Vector2D(temp.x - lightSeg.end1X, temp.y - lightSeg.end1Y);
					if(refClone.length() <= maxReflect) {
						reflect.x = temp.x;
						reflect.y = temp.y;
						toReflect.x = lightSeg.end2X - reflect.x;
						toReflect.y = lightSeg.end2Y - reflect.y;
						mirror = curMirror.mirrorSeg;
						reflected = true;
						maxReflect = refClone.length();			
					}
				}
				
			}
			
			if(reflected == true){
				double theta = mirror.getTheta();
				
				double mirNormal = Math.atan2(mirror.end1X - mirror.end2X, mirror.end1Y - mirror.end2Y);
				double reflectTheta = lightSeg.getTheta() + (mirNormal - lightSeg.getTheta()) * 2;
				double reflectLength = lightSeg.getLength() - maxReflect;
				
				Vector2D reflection = new Vector2D((Math.cos(2*theta)*toReflect.x + Math.sin(2*theta)*toReflect.y),(Math.sin(2*theta)*toReflect.x + -1 * Math.cos(2*theta)*toReflect.y));			
				//Vector2D reflection = new Vector2D(Math.cos(reflectTheta) * reflectLength, Math.sin(reflectTheta) * reflectLength);
				// Possible bugfix...
				//if((theta < 0.01 && theta > -0.01) || mirror.getSlope() > 1.0 ) {
				/*
				if(theta < 0) {
					reflection.x = -reflection.x;
					reflection.y = -reflection.y;
				}
				*/
				//}
				
				
				
				Vector2D normReflect = new Vector2D(reflection.x, reflection.y);
				normReflect.normalize();
				
				LineSeg refLight = new LineSeg(reflect.x + normReflect.x, reflect.y + normReflect.y, reflect.x + reflection.x, reflect.y + reflection.y);
				
				
				
				double bLeng = 6000.0 / lightSeg.getLength() * (maxReflect / lightSeg.getLength());
				if(bLeng > 60)
					bLeng = 60;
				if(this.recurse > 0)
					bLeng = bLeng / (this.recurse * 2 + 2);
				LineSeg refBase = null;
				
				if(theta > 0) {//((reflectTheta + 2 * Math.PI) % Math.PI) > Math.PI / 2) {
					refBase = new LineSeg(reflect.x - (bLeng / 2) * Math.cos(theta), reflect.y - (bLeng / 2) * Math.sin(theta),
						reflect.x + (bLeng / 2) * Math.cos(theta), reflect.y + (bLeng / 2) * Math.sin(theta));
				} else {
					refBase = new LineSeg(reflect.x + (bLeng / 2) * Math.cos(theta), reflect.y + (bLeng / 2) * Math.sin(theta),
							reflect.x - (bLeng / 2) * Math.cos(theta), reflect.y - (bLeng / 2) * Math.sin(theta));
				}
				
				
				int bounceNum = this.recurse + 1;
				//System.out.println(bounceNum);
				
				Light reflectedLight = new Light(refLight, refBase, bounceNum);
				reflectedLight.reflect();
			}
			return reflected;
		}
		return false;
	}
	
	public void Paint (Graphics2D g2D) {
	
		double bLeng;
		double aLeng = lightSeg.getLength();
		double bXLen;
		double bYLen;	
		
		LineSeg refBase = null;
		
		maxDist = aLeng;
		bLeng = 6000.0 / aLeng;
		if(bLeng > 60)
			bLeng = 60;
		if(this.recurse > 0)
			bLeng = bLeng / (this.recurse * 2);
		
		//Vector2D shortestIntersect = null;
		for(Wall curWall : Wall.allWall) {
			Vector2D intersect = lightSeg.getIntersect(curWall.wallSeg);
			if(intersect != null) {
				Vector2D intClone = new Vector2D(intersect.x - lightSeg.end1X, intersect.y - lightSeg.end1Y);
				if(intClone.length() < maxDist) {
					maxDist = intClone.length();
					bLeng = 6000.0 / aLeng * (maxDist / aLeng);
					if(bLeng > 60)
						bLeng = 60;
					if(this.recurse > 0)
						bLeng = bLeng / (this.recurse * 2);
					
					double theta = curWall.wallSeg.getTheta();
					if(theta > 0) {
						refBase = new LineSeg(intersect.x - (bLeng / 2) * Math.cos(theta), intersect.y - (bLeng / 2) * Math.sin(theta),
							intersect.x + (bLeng / 2) * Math.cos(theta), intersect.y + (bLeng / 2) * Math.sin(theta));
					} else {
						refBase = new LineSeg(intersect.x + (bLeng / 2) * Math.cos(theta), intersect.y + (bLeng / 2) * Math.sin(theta),
								intersect.x - (bLeng / 2) * Math.cos(theta), intersect.y - (bLeng / 2) * Math.sin(theta));
					}
					//shortestIntersect = intersect;
				}
			}
		}
		
		for(Mirror curMirror : Mirror.allMirror) {
			Vector2D intersect = lightSeg.getIntersect(curMirror.mirrorSeg);
			if(intersect != null) {
				Vector2D intClone = new Vector2D(intersect.x - lightSeg.end1X, intersect.y - lightSeg.end1Y);
				if(intClone.length() < maxDist) {
					maxDist = intClone.length();
					bLeng = 6000.0 / aLeng * (maxDist / aLeng);
					if(bLeng > 60)
						bLeng = 60;
					if(recurse > 0)
						bLeng = bLeng / (recurse * 2);
					
					double theta = curMirror.mirrorSeg.getTheta();
					if(theta > 0) {
						refBase = new LineSeg(intersect.x - (bLeng / 2) * Math.cos(theta), intersect.y - (bLeng / 2) * Math.sin(theta),
							intersect.x + (bLeng / 2) * Math.cos(theta), intersect.y + (bLeng / 2) * Math.sin(theta));
					} else {
						refBase = new LineSeg(intersect.x + (bLeng / 2) * Math.cos(theta), intersect.y + (bLeng / 2) * Math.sin(theta),
								intersect.x - (bLeng / 2) * Math.cos(theta), intersect.y - (bLeng / 2) * Math.sin(theta));
					}
					//shortestIntersect = intersect;
				}
			}
		}
		
		
		
		
		bYLen = Math.sqrt((bLeng * bLeng)/(1 + (Math.pow(((lightSeg.end2Y-lightSeg.end1Y)/(lightSeg.end2X-lightSeg.end1X)),2))));
		bXLen = -1 * bYLen * ((lightSeg.end2Y-lightSeg.end1Y)/(lightSeg.end2X-lightSeg.end1X));
		
		/*
		Point leftSegEnd = new Point((int)(lightSeg.end2X + bXLen), (int)(lightSeg.end2Y + bYLen));
		Point rightSegEnd = new Point((int)(lightSeg.end2X - bXLen), (int)(lightSeg.end2Y - bYLen));
		
		int[] xPoints = {(int)lightSeg.end1X, (int)leftSegEnd.x, (int)rightSegEnd.x};
		int[] yPoints = {(int)lightSeg.end1Y, (int)leftSegEnd.y, (int)rightSegEnd.y};
		*/
		
		//Polygon light = new Polygon(xPoints, yPoints, 3);
		/*
		g2D.setColor(new Color(0, 0, 0, 64));
		g2D.fillPolygon(light);
		g2D.setColor(new Color(0, 0, 0, 128));
		g2D.drawPolygon(light);
		*/
		
		double lt = lightSeg.getTheta();
		double ll = lightSeg.getLength();
		
		double scStart = maxDist / aLeng; 
		
		for(double sc = scStart; sc > 0.1; sc-=0.03) {
			Point leftSegEnd = null;
			Point rightSegEnd = null;
			
			if(sc == scStart && refBase != null) {
				leftSegEnd = new Point((int) refBase.end1X, (int) refBase.end1Y);
				rightSegEnd = new Point((int) refBase.end2X, (int) refBase.end2Y);
			} else {
				leftSegEnd = new Point((int)(lightSeg.end2X + bXLen * sc * sc - Math.cos(lt) * (1.0 - sc) * ll), (int)(lightSeg.end2Y + bYLen * sc * sc - Math.sin(lt) * (1.0 - sc) * ll));
				rightSegEnd = new Point((int)(lightSeg.end2X - bXLen * sc * sc - Math.cos(lt) * (1.0 - sc) * ll), (int)(lightSeg.end2Y - bYLen * sc * sc - Math.sin(lt) * (1.0 - sc) * ll));
			}
			//LineSeg curLightSeg = new LineSeg(lightSeg.end1X, lightSeg.end1Y, leftSegEnd.x, leftSegEnd.y);
			
			int[] xPoints = {(int) baseSeg.end1X, (int) baseSeg.end2X, (int)leftSegEnd.x, (int)rightSegEnd.x};
			int[] yPoints = {(int) baseSeg.end1Y, (int) baseSeg.end2Y, (int)leftSegEnd.y, (int)rightSegEnd.y};
			
			Polygon light = new Polygon(xPoints, yPoints, 4);
			
			g2D.setColor(new Color(0, 0, 0, ((int) (sc * 255))));
			
			g2D.fillPolygon(light);
			g2D.setColor(new Color(0, 0, 0, 250));
			g2D.drawPolygon(light);
			
		}
		
		leftLightSeg = new LineSeg(lightSeg.end1X, lightSeg.end1Y, 
				(lightSeg.end2X + bXLen * scStart * scStart - Math.cos(lt) * (1.0 - scStart) * ll),
				(lightSeg.end2Y + bYLen * scStart * scStart - Math.sin(lt) * (1.0 - scStart) * ll));
		
		rightLightSeg = new LineSeg(lightSeg.end1X, lightSeg.end1Y,
				(lightSeg.end2X - bXLen * scStart * scStart - Math.cos(lt) * (1.0 - scStart) * ll),
				(lightSeg.end2Y - bYLen * scStart * scStart - Math.sin(lt) * (1.0 - scStart) * ll));
	}
	
}
