package main;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

import javax.swing.*;
import javax.swing.Timer;

import java.awt.image.*;
import java.awt.event.*;

import java.util.Calendar;

public class RenderFrame extends JFrame implements Runnable, MouseMotionListener, MouseListener, KeyListener {
	int sizeX, sizeY; // Respectively, refresh rate, and window size in pixels
	BufferedImage backBuffer; // Image buffer for smooth rendering
	BufferedImage lightBuffer; // Image buffer for lighting
	
	Font ingameFont;
	static Image lightIm, bg, mote, mirror, spotlight;
	static Image baseMask, mirrorMask;
	static ArrayList<Image> baseIm;
	static BufferedImage WallTexture;
	static RenderFrame pubFrame;
	static Timer timer;
	int mouseX, mouseY;
	ArrayList<DustCloud> clouds;	
	double rotator;
	static int phase; //0 for day, 1 for night
	static Calendar time = Calendar.getInstance();
	int rounds;
	
	boolean renderLight, renderWall, renderDust, renderBlur;
	Light mainLight;
	Base mainBase;
	
	double playerX, playerY;
	boolean gameStart;
	
	RenderFrame(int sizeX, int sizeY) {
		super("A Study in Dust");
		this.sizeX = sizeX;
		this.sizeY = sizeY;
		this.setSize(new Dimension(sizeX, sizeY));
		this.setLocation(10, 10);
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		backBuffer = new BufferedImage(sizeX, sizeY, BufferedImage.TYPE_INT_ARGB);
		lightBuffer = new BufferedImage(sizeX, sizeY, BufferedImage.TYPE_INT_ARGB);
		this.addMouseMotionListener(this);
		this.addMouseListener(this);
		this.addKeyListener(this);
		this.setResizable(false);
		
		rounds = 0;
		
		lightIm = (new ImageIcon("img/light.png")).getImage();
		bg = (new ImageIcon("img/bg.png")).getImage();
		mote = (new ImageIcon("img/mote.png")).getImage();
		spotlight = (new ImageIcon("img/spotlight.PNG")).getImage();
		mirror = (new ImageIcon("img/mirror.png")).getImage(); //get the image for the mirror
		baseIm = new ArrayList<Image>();
		for(int i = 1; i < 11; i++) {
			baseIm.add(new ImageIcon("img/base_" + i + ".png").getImage());
		}
		baseMask = (new ImageIcon("img/base_mask.png")).getImage();
		mirrorMask = (new ImageIcon("img/mirror_mask.png")).getImage();
		ingameFont = new Font("Arial", Font.PLAIN, 22);
		
		clouds = new ArrayList<DustCloud>();
		
		clouds.add(new DustCloud(20, 400, 100, 100, 0.2, 0.2, 10.0, 1.5, 1.0));
		clouds.add(new DustCloud(20, 200, 100, 100, 0.2, 0.2, 1.0, 2.0, 4.0));
		clouds.add(new DustCloud(20, 600, 100, 100, 0.2, 0.2, 1.0, 6.0, 1.5));
		
		for(int i = 0; i < 3; i++) {
			addDustCloud();
		}
		
		WallTexture = new BufferedImage(32, 32, BufferedImage.TYPE_INT_ARGB);//new BufferedImage((new ImageIcon("img/WallTexture.png")).getImage());
		WallTexture.getGraphics().drawImage((new ImageIcon("img/WallTexture.png")).getImage(), 0, 0, this);
		rotator = 0;
		
		Wall wall1 = new Wall(new LineSeg(200, 200, 200, 300));
		Wall wall2 = new Wall(new LineSeg(200, 300, 300, 300));
		Wall wall3 = new Wall(new LineSeg(300, 300, 300, 300));
		Wall wall4 = new Wall(new LineSeg(300, 300, 200, 200));
		
		Wall wall5 = new Wall(new LineSeg(500, 400, 500, 300));
		Wall wall6 = new Wall(new LineSeg(500, 300, 550, 300));
		Wall wall7 = new Wall(new LineSeg(500, 400, 650, 400));
		Wall wall8 = new Wall(new LineSeg(550, 300, 650, 400));
		
		Wall wall9 = new Wall(new LineSeg(50, 100, 50, 50));
		Wall wall10 = new Wall(new LineSeg(50, 50, 100, 50));
		Wall wall11 = new Wall(new LineSeg(50, 100, 200, 100));
		Wall wall12 = new Wall(new LineSeg(100, 50, 200, 100));
			
		Wall wall15 = new Wall(new LineSeg(150, 400, 150, 500));
		Wall wall16 = new Wall(new LineSeg(150, 500, 250, 500));
		Wall wall17 = new Wall(new LineSeg(250, 500, 250, 500));
		Wall wall18 = new Wall(new LineSeg(250, 500, 150, 400));
		
		mainLight = new Light(new LineSeg(400.0, 300.0, 0.0, 0.0), 0);
		
//		for(int i = 0; i < 100; i++) {
//			Mirror curMirror = new Mirror((int) (Math.random() * 800), (int) (Math.random() * 600));
//			curMirror.setAngle(Math.random() * Math.PI);
//		}
		
		mainBase = new Base(400, 300);
		
		playerX = 400.0;
		playerY = 300.0;
		
		renderLight = true; //Render lights/shadow layer
		renderWall = true; //Render obstacles/walls
		renderDust = true; //Render dust particles
		renderBlur = true; //Render motion blur
		
		phase = 0; //day at the beginning		
		time = Calendar.getInstance();
		gameStart = false;
	}
	
	public static void main(String[] args) {
		startTheGame();
		//Create the opening frame
		final JFrame overall = new JFrame("A Study in Dust");
		overall.setSize(800, 600);
		JLayeredPane openingFrame = new JLayeredPane();
		openingFrame.setLayout(null);
		openingFrame.setSize(800, 600);
		JButton startGame = new JButton("Start the Game");
		startGame.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				
				RenderFrame.pubFrame.gameStart = true;
				RenderFrame.pubFrame.setVisible(true);
				
				overall.dispose();
			/*
			// Create a default frame
			RenderFrame.pubFrame = new RenderFrame(800, 600);
			// Start this frame rendering...
			RenderFrame.pubFrame.run();
			// and draw it once, just to be safe.
			RenderFrame.pubFrame.repaint();
//			startTheGame(); 
			 */
			
			}
		});
		Dimension size = startGame.getPreferredSize();
		startGame.setBounds(400 - size.width/2, 500, size.width, size.height);
		
		//the background
		ImageIcon background = new ImageIcon("img/bg.png");
		JLabel fake = new JLabel("", background, JLabel.CENTER);
		size = fake.getPreferredSize();
		fake.setBounds(0, 0, 800, 600);
		
		//the logo
		ImageIcon logo = new ImageIcon("img/logo.gif");
		JLabel fake2 = new JLabel("", logo, JLabel.CENTER);
		size = fake2.getPreferredSize();
		fake2.setBounds(150, 30, 500, 110);
		
		//the names
		JLabel created = new JLabel("<html><font size=5 color = red>Created By</font></html>", JLabel.CENTER);
		size = created.getPreferredSize();
		created.setBounds(350, 250, size.width, size.height);
		
		JLabel garon = new JLabel("<html><font size=4 color = red>Garon Clements</font></html>", JLabel.CENTER);
		size = garon.getPreferredSize();
		garon.setBounds(345, 280, size.width, size.height);
		
		JLabel chris = new JLabel("<html><font size=4 color = red>Chris Drouin</font></html>", JLabel.CENTER);
		size = chris.getPreferredSize();
		chris.setBounds(355, 300, size.width, size.height);
		
		JLabel evan = new JLabel("<html><font size=4 color = red>Evan May</font></html>", JLabel.CENTER);
		size = evan.getPreferredSize();
		evan.setBounds(363, 320, size.width, size.height);
		
		openingFrame.add(fake, new Integer(0));
		openingFrame.add(fake2, new Integer(1));
		openingFrame.add(startGame, new Integer(2));
		openingFrame.add(created, new Integer(2));
		openingFrame.add(garon, new Integer(2));
		openingFrame.add(chris, new Integer(2));
		openingFrame.add(evan, new Integer(2));
		openingFrame.setVisible(true);
		overall.add(openingFrame);
		overall.setVisible(true);
		
	}
	
	public static void startTheGame()
	{
		// Create a default frame
		pubFrame = new RenderFrame(800, 600);
		// Start this frame rendering...
		pubFrame.run();
		// and draw it once, just to be safe.
		pubFrame.repaint();
	}
	
	public void mouseDragged(MouseEvent e) {

	}
	
	public void mouseExited(MouseEvent e) {

	}
	
	public void mouseEntered(MouseEvent e) {

	}
	
	public void mouseClicked(MouseEvent e) {
		if (RenderFrame.phase == 0)
		{
			new Mirror(e.getX(), e.getY());
			Player.num_Mirrors--; //place a mirror if it is day
			RenderFrame.time = Calendar.getInstance();
		}
		
	}
	
	public void mousePressed(MouseEvent e) {
		
	}
	
	public void mouseReleased(MouseEvent e) {
		
	}
	
	public void mouseMoved(MouseEvent e) {
		Point pos = e.getPoint();
		mouseX = pos.x;
		mouseY = pos.y;
	}
	
	public void run() {
		//this.setVisible(true);
		RenderThread renderThis = new RenderThread(this);
		renderThis.start();
	}
	
	public void addDustCloud() {
		double radius = Math.random() * 400 + 500;
		double theta = Math.random() * Math.PI * 2 - Math.PI;
		clouds.add(new DustCloud(20, radius * Math.cos(theta), radius * Math.sin(theta), 100, Math.random() * 2.0 - 1, Math.random() * 2.0 - 1, Math.random() * 10.0 + 0.5, Math.random() * 10.0 + 0.5,Math.random() * 10.0 + 0.5));
	}
	
	public void paint (Graphics g) {
		//super.paint(g);
		//Graphics2D g2D = (Graphics2D) g;
		
		Graphics2D g2D = (Graphics2D) backBuffer.getGraphics();
		Graphics2D light = (Graphics2D) lightBuffer.getGraphics();
		
		// Set light so that source alpha replaces existing alpha
		light.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_IN, 1.0f));
		// Set everything dark...
		light.setBackground(new Color(0, 0, 0, 240));
		light.clearRect(0, 0, sizeX, sizeY);
		// and now punch a large whole in that darkness
		
		mainLight.lightSeg.end1X = playerX;
		mainLight.lightSeg.end1Y = playerY;
		mainLight.baseSeg.end1X = playerX;
		mainLight.baseSeg.end2X = playerX;
		mainLight.baseSeg.end1Y = playerY;
		mainLight.baseSeg.end2Y = playerY;
		
		mainLight.lightSeg.end2X = mouseX;
		mainLight.lightSeg.end2Y = mouseY;
		mainLight.reflectChecked = false;
		
		Light.allLight.clear();
		Light.allLight.add(mainLight);
		
		mainLight.reflect();
		
		if(renderLight) {
			for(int i = 0; i < Light.allLight.size(); i++) {
				Light curLight = Light.allLight.get(i);
				curLight.Paint(light);				
			}
		}
		
		g2D.drawImage(bg, 0, 0, this);
		if(renderWall) {
			for(Wall curWall : Wall.allWall) {
				curWall.paint(g2D);
				if(renderLight)
					curWall.lightPaint(light);
			}
		}
		mainBase.paint(g2D);
		light.drawImage(baseMask, mainBase.posX - 32, mainBase.posY - 32, this);
		
		for (Mirror curMirror : Mirror.allMirror)
		{
			curMirror.setAngle(curMirror.mirrorSeg.getTheta() + 0.01);
			
			curMirror.paint(g2D);
			if(renderLight) 
				curMirror.lightPaint(light);
		}
		
		g2D.setPaint(null);
		//g2D.setColor(Color.BLUE);
		g2D.drawImage(spotlight, (int) playerX - 16, (int) playerY - 16, this);
		
		
		//Render light buffer onto the primary buffer
		if(renderLight && phase == 1)
		{
			g2D.drawImage(lightBuffer, 0, 0, this);
			g2D.setColor(Color.GRAY);
			Calendar temp = Calendar.getInstance();
			int difference = RenderFrame.time.get(RenderFrame.time.SECOND) - temp.get(temp.SECOND);
			if (difference <= 0) {
				difference *= -1;
			} else {
				difference = 60 - difference; 
			}
			
			g2D.setFont(ingameFont);
			g2D.drawString("Time left until day: " + 
					Integer.toString(59 - difference) + " seconds", sizeX - 300, sizeY - 10);
				
		}
		
		if(renderDust && phase == 1) {
			for(DustCloud cloud : clouds) {
				cloud.paint(g2D);
			}
		}
		
		
		
		Graphics2D force2D = (Graphics2D) g;
		
		if(renderBlur) {
			force2D.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f));
		} else {
			force2D.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1.0f));
		}
		force2D.drawImage(backBuffer, 0, 0, this);
		
	}
	
	public void drawImageRotScale(Graphics2D g2D, Image toDraw, int x, int y, double scaleX, double scaleY, double rot) {
		AffineTransform imTransform = new AffineTransform();
		double theta = Math.atan2(y, x) - rot;
		double radius = Math.sqrt(x * x + y * y);
		imTransform.rotate(rot, 0.0, toDraw.getHeight(this) / 2 * scaleY);
		imTransform.scale(scaleX, scaleY);
		imTransform.translate(radius * Math.cos(theta) / scaleX, radius * Math.sin(theta) / scaleY);
		g2D.drawImage(toDraw, imTransform, this);
		
		
	}

	public void keyTyped(KeyEvent e) {
		
	}

	static double minRepel = 0.2;
	static double playerVel = 1.5;
	public void keyPressed(KeyEvent e) {
		int code = e.getKeyCode();
		Dust fake = new Dust(playerX, playerY, 0, 0);
		Vector2D repel = new Vector2D(0, 0);
		for(Wall curWall : Wall.allWall) {
			repel.add(curWall.wallSeg.getRepulsion(fake));
		}
		if(code == KeyEvent.VK_W) {
			playerY -= playerVel;
			Vector2D repel2 = new Vector2D(0, 0);
			fake.vel.y = -(playerVel / 10);
			for(Wall curWall : Wall.allWall) {
				repel2.add(curWall.wallSeg.getRepulsion(fake));
			}
			if(repel2.length() > minRepel) {
				repel2.scale(0.8);
				if(repel2.length() > playerVel * 2)
					repel2.scale(0.1);
				playerX += repel2.x;
				playerY += repel2.y;
			}
		} else if(code == KeyEvent.VK_S) {
			playerY += playerVel;
			fake.vel.y = playerVel / 10;
			Vector2D repel2 = new Vector2D(0, 0);
			for(Wall curWall : Wall.allWall) {
				repel2.add(curWall.wallSeg.getRepulsion(fake));
			}
			if(repel2.length() > minRepel) {
				repel2.scale(0.8);
				if(repel2.length() > playerVel * 2)
					repel2.scale(0.1);
				playerX += repel2.x;
				playerY += repel2.y;
			}
		} else if(code == KeyEvent.VK_A) {
			playerX -= playerVel;
			fake.vel.x = -(playerVel / 10);
			Vector2D repel2 = new Vector2D(0, 0);
			for(Wall curWall : Wall.allWall) {
				repel2.add(curWall.wallSeg.getRepulsion(fake));
			}
			if(repel2.length() > minRepel) {
				repel2.scale(0.8);
				if(repel2.length() > playerVel * 2)
					repel2.scale(0.1);
				playerX += repel2.x;
				playerY += repel2.y;
			}
		} else if(code == KeyEvent.VK_D) {
			playerX += playerVel;
			fake.vel.x = playerVel / 10;
			Vector2D repel2 = new Vector2D(0, 0);
			for(Wall curWall : Wall.allWall) {
				repel2.add(curWall.wallSeg.getRepulsion(fake));
			}
			if(repel2.length() > minRepel) {
				repel2.scale(0.8);
				if(repel2.length() > playerVel * 2)
					repel2.scale(0.1);
				playerX += repel2.x;
				playerY += repel2.y;
			}
		}
		
	}

	public void keyReleased(KeyEvent e) {
		int code = e.getKeyCode();
		if(code == KeyEvent.VK_F1) {
			renderLight = !renderLight;
		} else if(code == KeyEvent.VK_F2) {
			renderWall = !renderWall;
		} else if(code == KeyEvent.VK_F3) {
			renderDust = !renderDust;
		} else if(code == KeyEvent.VK_F4) {
			renderBlur = !renderBlur;
		}
	}
}

class RenderThread extends Thread {
	RenderFrame sourceFrame;
	
	/**
	 * Standard constructor for this rendering thread
	 * @param sourceFrame
	 */
	RenderThread (RenderFrame sourceFrame) {
		this.sourceFrame = sourceFrame;
	}
	
	/**
	 * Method that starts the thread rendering.
	 */
	public void run () {
		while(true) {
			try {
				// Hard-coded 60hz delay; change as necessary
				sleep (1000 / 60);
			} catch (InterruptedException e) {
				// Don't do anything, Java just wants this to be caught. -_-
			}
			if(sourceFrame.gameStart) {
				sourceFrame.validate();
				sourceFrame.repaint();
				
				for(DustCloud cloud : sourceFrame.clouds) {
					cloud.moveCloud(sourceFrame.mouseX, sourceFrame.mouseY);
					
					sourceFrame.mainBase.damageHealth();
					sourceFrame.mainBase.damagePoints = 0.0;
				}
				Calendar temp = Calendar.getInstance();
				int difference = RenderFrame.time.get(RenderFrame.time.SECOND) - temp.get(temp.SECOND);
				if (difference <= 0)
					difference *= -1;
				else
					difference = 60 - difference; 
				if (difference >= 19 && RenderFrame.phase == 1)
				{
					RenderFrame.phase = 0;
	//				RenderFrame.time = Calendar.getInstance();
					Player.num_Mirrors++;
				}
			}
					
			if (Player.getNumMirrors() == 0 && RenderFrame.phase != 1)
			{
				RenderFrame.phase = 1; //player has no mirrors, so it becomes night
				RenderFrame.pubFrame.rounds++;
				for(int i = 0; i < RenderFrame.pubFrame.rounds + 4; i++) {
					sourceFrame.addDustCloud();
				}
				
			}
			Calendar temp = Calendar.getInstance();
			int difference = RenderFrame.time.get(RenderFrame.time.SECOND) - temp.get(temp.SECOND);
			if (difference <= 0)
				difference *= -1;
			else
				difference = 60 - difference; 
			if (difference >= 59 && RenderFrame.phase == 1)
			{
				RenderFrame.phase = 0;
//				RenderFrame.time = Calendar.getInstance();
				Player.num_Mirrors++;
				sourceFrame.clouds.clear();
				if(sourceFrame.mainBase.hitPoints < 100.0)
					sourceFrame.mainBase.hitPoints+=25.0;
			}
			if(RenderFrame.pubFrame.mainBase.hitPoints < 1.0 && RenderFrame.pubFrame.isVisible()) {
				new GameOverFrame(RenderFrame.pubFrame.rounds);
				RenderFrame.pubFrame.setVisible(false);
			}
			
			
		}
	}
}

class GameOverFrame extends JFrame {
	static GameOverFrame instance;
	GameOverFrame(int rounds) {
		super("Game Over");
		instance = this;
		this.setSize(340, 240);
		this.setLocation(200, 200);
		Font GameFont = new Font("Arial", Font.BOLD, 50);
		JLabel GameOverLabel = new JLabel("Game Over");
		GameOverLabel.setFont(GameFont);
		JPanel contentPane = new JPanel();
		this.setContentPane(contentPane);
		contentPane.add(GameOverLabel);
		JLabel RoundsLabel = new JLabel("You lasted " + rounds + " rounds against the dust - thanks for playing!");
		contentPane.add(RoundsLabel);
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		JButton endButton = new JButton("Exit");
		endButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				instance.setVisible(false);
			}
		});
		contentPane.add(endButton);
		this.setVisible(true);
	}
}