package main;

import java.awt.*;
import java.util.ArrayList;

public class Wall {
	static ArrayList<Wall> allWall = new ArrayList<Wall>();
	
	LineSeg wallSeg; //Representative line segment
	Wall(LineSeg wallSeg) {
		this.wallSeg = wallSeg;
		allWall.add(this);
	}
	
	public void paint(Graphics2D g2D) {
		
		g2D.setColor(Color.BLACK);
		//g2D.setPaint(new GradientPaint(-20.0f * (float) Math.cos(wallSeg.getTheta() + Math.PI), -20.0f * (float) Math.sin(wallSeg.getTheta() + Math.PI), Color.BLACK, 20.0f * (float) Math.cos(wallSeg.getTheta() + Math.PI), 20.0f * (float) Math.sin(wallSeg.getTheta() + Math.PI), Color.DARK_GRAY));
				
		g2D.setStroke(new BasicStroke(15.0f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
		g2D.drawLine((int) wallSeg.end1X, (int) wallSeg.end1Y, (int) wallSeg.end2X, (int) wallSeg.end2Y);
		
		
		g2D.setColor(Color.DARK_GRAY);
		g2D.setStroke(new BasicStroke(13.0f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
		g2D.drawLine((int) wallSeg.end1X, (int) wallSeg.end1Y, (int) wallSeg.end2X, (int) wallSeg.end2Y);
		
		
		g2D.setPaint(new TexturePaint(RenderFrame.WallTexture, new Rectangle(32, 32)));
		g2D.setStroke(new BasicStroke(10.0f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
		g2D.drawLine((int) wallSeg.end1X, (int) wallSeg.end1Y, (int) wallSeg.end2X, (int) wallSeg.end2Y);
		
	}
	
	public void lightPaint(Graphics2D g2D) {
		g2D.setColor(new Color(0, 0, 0, 130));
		g2D.setStroke(new BasicStroke(15.0f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
		g2D.drawLine((int) wallSeg.end1X, (int) wallSeg.end1Y, (int) wallSeg.end2X, (int) wallSeg.end2Y);
	}
}

class LineSeg {
	double end1X, end2X, end1Y, end2Y;
	
	/**
	 * Basic line segment constructor
	 * @param end1X end 1 x coord
	 * @param end1Y end 1 y coord
	 * @param end2X end 2 x coord
	 * @param end2Y end 2 y coord
	 */
	public LineSeg(double end1X, double end1Y, double end2X, double end2Y) {
		this.end1X = end1X;
		this.end1Y = end1Y;
		this.end2X = end2X;
		this.end2Y = end2Y;
	}
	
	/**
	 * Finds the slope of the line segment
	 * @return slope (or NaN if slope is infinite)
	 */
	public double getSlope() {
		if(Math.abs(end1X - end2X) > 0.01) {
			return (end1Y - end2Y) / (end1X - end2X);
		} else {
			return 50001;
		}
	}
	
	/**
	 * Returns the length of the line segment in pixels
	 * @return length of the line segment
	 */
	public double getLength() {
		return Math.sqrt((end1X - end2X) * (end1X - end2X) + (end1Y - end2Y) * (end1Y - end2Y));
	}
	
	/**
	 * Angle the line segment forms with x-axis
	 * @return angle (in radians)
	 */
	public double getTheta() {
		return Math.atan2(end2Y - end1Y, end2X - end1X);
	}
	
	/**
	 * y-coordinate at which the line [would] intersects the Y axis
	 * @return the 
	 */
	public double getYIntersect() {
		if(getSlope() == Double.NaN)
			return Double.NaN;
		return (getSlope() * -end1X + end1Y);
	}
	
	/**
	 * Return a vector for a given dustmote's repulsion from this line segment
	 * @param mote to find repulsion
	 * @return repulsion vector
	 */
	public Vector2D getRepulsion(Dust mote) {
		Vector2D repulse = null;
		double slope = getSlope();
		double invSlope;
		double bounceAngle = 5.0;
		
		if(slope == Double.NaN) {
			//invSlope = 0.0;
			if(end1Y < end2Y) {
				if(mote.pos.y >= end1Y && mote.pos.y <= end2Y) {
					repulse = new Vector2D(mote.pos.x - end1X, 0.0);
				} else if(mote.pos.y < end1Y) {
					repulse = new Vector2D(mote.pos.x - end1X, mote.pos.y - end1Y);
				} else {
					repulse = new Vector2D(mote.pos.x - end1X, mote.pos.y - end2Y);
				}
			} else {
				if(mote.pos.y >= end2Y && mote.pos.y <= end1Y) {
					repulse = new Vector2D(mote.pos.x - end1X, 0.0);
				} else if(mote.pos.y < end2Y) {
					repulse = new Vector2D(mote.pos.x - end1X, mote.pos.y - end2Y);
				} else {
					repulse = new Vector2D(mote.pos.x - end1X, mote.pos.y - end1Y);
				}
			}
		} else if(slope <= 0.01 && slope >= -0.01) {
			if(end1X < end2X) {
				if(mote.pos.x >= end1X && mote.pos.x <= end2X) {
					repulse = new Vector2D(0.0, mote.pos.y - end1Y);
					repulse.scale(0.75);
				} else if(mote.pos.x < end1X) {
					repulse = new Vector2D(mote.pos.x - end1X, mote.pos.y - end1Y);
					repulse.scale(40.0);
				} else {
					repulse = new Vector2D(mote.pos.x - end1X, mote.pos.y - end2Y);
					repulse.scale(40.0);
				}
			} else {
				if(mote.pos.x >= end2X && mote.pos.x <= end1X) {
					repulse = new Vector2D(0.0, mote.pos.y - end2Y);
					repulse.scale(0.75);
				} else if(mote.pos.x < end2X) {
					repulse = new Vector2D(mote.pos.x - end1X, mote.pos.y - end2Y);
					repulse.scale(40.0);
				} else {
					repulse = new Vector2D(mote.pos.x - end1X, mote.pos.y - end1Y);
					repulse.scale(40.0);
				}
			}
		} else {
			invSlope = -(1.0 / slope);
			double moteIntersect = (invSlope * -mote.pos.x + mote.pos.y);
			double x = (moteIntersect - getYIntersect()) / (slope - invSlope);
			double y = slope * x + getYIntersect();
			
			mote.intersect = new Vector2D(x, y);
			
			boolean in1 = false;
			if(end2X > end1X) {
				if(x >= end1X && x <= end2X)
					in1 = true;
			} else {
				if(x <= end1X && x >= end2X)
					in1 = true;
			}
			
			if(in1) {
				repulse = new Vector2D(mote.pos.x - x, mote.pos.y - y);
			} else {
				// More direct repulsion from ends
				bounceAngle = 20.0;
				if(end2X > end1X) {
					if(x < end1X) {
						repulse = new Vector2D(mote.pos.x - end1X, mote.pos.y - end1Y);
						repulse.scale(40.0);
					} else {
						repulse = new Vector2D(mote.pos.x - end2X, mote.pos.y - end2Y);
						repulse.scale(40.0);
					}
				} else {
					if(x < end2X) {
						repulse = new Vector2D(mote.pos.x - end2X, mote.pos.y - end2Y);
						repulse.scale(40.0);
					} else {
						repulse = new Vector2D(mote.pos.x - end1X, mote.pos.y - end1Y);
						repulse.scale(40.0);
					}
				}
				
			}
			
		}
		
		
		if(Math.abs(mote.vel.theta() - repulse.theta()) > 3.1415) {
			if(mote.vel.theta() < repulse.theta()) {
				repulse.rotate(Math.PI / bounceAngle);
			} else {
				repulse.rotate(-Math.PI / bounceAngle);
			}
		} else {
			if(mote.vel.theta() < repulse.theta()) {
				repulse.rotate(-Math.PI / bounceAngle);
			} else {
				repulse.rotate(Math.PI / bounceAngle);
			}
		}
		
		if(repulse.length() < 75.0) {
			repulse.x = 1.0 / repulse.x;
			repulse.y = 1.0 / repulse.y;
	
		} else {
			repulse = new Vector2D(0.0, 0.0);
		}
		
		return repulse;
	}
	
	/**
	 * For debug only - draws the line segment in bright green
	 * @param g2D
	 */
	public void paint(Graphics2D g2D) {
		g2D.setColor(Color.GREEN);
		g2D.drawLine((int) end1X, (int) end1Y, (int) end2X, (int) end2Y);
	}
	
	/**
	 * Find the point at which two line segments intersect
	 * @param toCheck line segment to check against
	 * @return point at which the segments intersect (null if none)
	 */
	public Vector2D getIntersect (LineSeg toCheck) {
		double slope1 = getSlope();
		double slope2 = toCheck.getSlope();
		
		if(slope1 == Double.NaN && slope2 == Double.NaN) {
			// Can't collide if they're both vertical
			return null;
		} else if (Math.abs(slope1) > 5000) {
			// Check for collision if line 1 is vertical
			double y = end1X * toCheck.getSlope() + toCheck.getYIntersect();
			
			if(end2Y > end1Y) {
				if(y >= end1Y && y <= end2Y)
					return new Vector2D(end1X, y);
			} else {
				if(y <= end1Y && y >= end2Y)
					return new Vector2D(end1X, y);
			}
			return null;
		} else if (Math.abs(slope2) > 5000) {
			// Check for collision if line 2 is vertical
			double y = toCheck.end1X * getSlope() + getYIntersect();
			
			if(toCheck.end2Y > toCheck.end1Y) {
				if(y >= toCheck.end1Y && y <= toCheck.end2Y)
					return new Vector2D(toCheck.end1X, y);
			} else {
				if(y <= toCheck.end1Y && y >= toCheck.end2Y)
					return new Vector2D(toCheck.end1X, y);
			}
			return null;
		} else if(slope1 == slope2) {
			// Parallel, so no collision
			return null;
		} else {
			double x = (toCheck.getYIntersect() - getYIntersect()) / (slope1 - slope2);
			boolean in1 = false;
			boolean in2 = false;
			if(end2X > end1X) {
				if(x >= end1X && x <= end2X)
					in1 = true;
			} else {
				if(x <= end1X && x >= end2X)
					in1 = true;
			}
			if(toCheck.end2X > toCheck.end1X) {
				if(x >= toCheck.end1X && x <= toCheck.end2X)
					in2 = true;
			} else {
				if(x <= toCheck.end1X && x >= toCheck.end2X)
					in2 = true;
			}
			// Check for normal collision
			if(in1 && in2)
				return new Vector2D(x, x * slope1 + getYIntersect());
		}
		// No collision
		return null;
	}
}
